
#include <Application.h>
#include <InterfaceKit.h>
#include <SupportDefs.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <StopWatch.h>
#include "fftproto.h"
#include "clear.h"

#define COLOR_MESSAGE			'COLR'
#define CONFIG_MESSAGE			'CONF'

plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"fft-clear",1, PLUGIN_IS_VISUAL,
	"Clear",
	"By Joe Wang.\n<joe@timewarp.net>\n",
	&GetAbout,
	NULL,
	&GetClearPlugin,
	&DestroyClearPlugin
	};

fft_visual_plugin_ops ops={
	{PLUGIN_VISUAL_MAGIC, PLUGIN_VISUAL_VERSION,
		&FileChange, NULL, &GetConfigure, NULL, NULL},
	NULL,
	ScreenModified,
//status_t op_fft_visual_init(); 		(*Init);
	NULL,
	NULL,
	ClearDirect
	};

static plugin_descriptor *plugs[]={
	&plugin,
	0
	};

plugin_descriptor **get_plugin_list(void) {
	return plugs;
	}


class Clear : public BLooper {
private:
	int32 file_changes;
	BRect bounds;
	color_space color;
	bool keeprunning, suspend, needs_update;
	BStopWatch *stopwatch;
	BHandler *handler;
	
public:
	int32 color_value;
	bigtime_t time;
	int32 file_change;

	Clear();
	~Clear();
	void Quit();
	bool QuitRequested();
	bool SetHandler(BHandler *);
	void MessageReceived(BMessage *mes);
	status_t RenderDirect(void *, const short *fft, int32 count);
	status_t Render8(int8 *, const short *fft, int32 count);
	status_t Render16(int16 *, const short *fft, int32 count);
	status_t Render32(int32 *, const short *fft, int32 count);
	status_t ScreenModified(BRect, color_space);
	void FileChange(char *);
	};

void GetAbout() {
	BAlert *alert=new BAlert("About Clear FFT",
		"Plugin for FFT Plugin.\n"
		"joe@timewarp.net\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
	}

BView *GetConfigure(void *data) {
	Clear *looper = (Clear *)data;
	BView *view=new BView(BRect(0,0,280,140),"Configuration",0,0);
	if(!looper->SetHandler(view)) {
		delete view;
		return NULL;
		}
	view->SetViewColor(40,40,100);
	char *string = (char *)malloc(32);
	sprintf(string, "%d", (int )looper->file_change);
	view->AddChild(
		new BTextControl(BRect(5,5,175,25), "File Change", "# of songs", string, NULL)
		);
	sprintf(string, "%d", (int )looper->time);
	view->AddChild(
		new BTextControl(BRect(5,25,175,45), "Time", "Milliseconds", string, NULL)
		);
	free(string);
	BMessage *msg=NULL;
	BColorControl *color=new BColorControl(BPoint(5, 50), B_CELLS_32x8, 0.0F,
		"Color Box");
	color->SetValue(looper->color_value);
	msg=new BMessage(COLOR_MESSAGE);
	msg->AddPointer("Color Box", (void *)color);
	color->SetMessage(msg);
	color->SetTarget(NULL, looper);
	view->AddChild(color);

	msg=new BMessage(CONFIG_MESSAGE);
	msg->AddPointer("Config View", view);
	BButton *button=new BButton(BRect(100,110,150,135), "Apply", "Apply", msg);
	button->SetTarget(NULL, looper);
	view->AddChild(button);
	return view;
	}

void *GetClearPlugin(void **data,const char *, const char *, uint32, plugin_info *) {
	*data=new Clear();
	return &ops;
	}

void DestroyClearPlugin(void*,void *data) {
	Clear *looper=(Clear *)data;
	looper->Lock();
	looper->Quit();
	}

status_t ClearDirect(void *data, void *bitmap, const short *buffer, const short *fft, int32 count, void*) {
	Clear *looper=(Clear *)data;
	return looper->RenderDirect(bitmap, fft, count);
	}

status_t ScreenModified(void *data, BRect bounds, color_space col, mini_screen mini, int32 screen) {
	Clear *looper=(Clear *)data;
	return looper->ScreenModified(bounds, col);
	}

void FileChange(void *data, char *name) {
	((Clear *)data)->FileChange(name);
	}

// ==============================================================================
Clear :: Clear() : BLooper() {

	time = 0;
	file_change = file_changes = 0;
	suspend = false;
	needs_update = false;
	color_value=0;
	stopwatch = new BStopWatch("timer", true);

	keeprunning=true;
	Run();
	}

Clear :: ~Clear() {
	delete stopwatch;
	}

bool Clear :: QuitRequested() {
	keeprunning=false;
	return false;
	}

void Clear::Quit() {
	keeprunning=false;
	if(handler) {
		BView *view = (BView *)handler;
		if(view->Window() && view->Window()->Lock()) view->Window()->Quit();
		}
	BLooper::Quit();
	}

bool Clear :: SetHandler(BHandler *hand) {
	if(!hand) {
		handler = NULL;
		return true;
		}
	if(handler) {
		BView *view = (BView *)handler;
		if(view->Window()) view->Window()->Show();
		return false;
		}
	handler=hand;
	return true;
	}

status_t Clear :: RenderDirect(void *bitmap, const short *fft, int32 count) {
	if(!keeprunning) return B_ERROR;
	if(time>0) {		//check the timer
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		else if(stopwatch->ElapsedTime()>time*1000) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	if(file_change>0) {
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		if(file_changes >= file_change) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	if(!needs_update) return B_OK;

	int32 col=(int32 )color;
	switch(col) {
	case B_CMAP8:
		return Render8((int8 *)bitmap, fft, count);
	case B_RGB15:
	case B_RGBA15:
	case B_RGB16:
		return Render16((int16 *)bitmap, fft, count);
	case B_RGB32:
		return Render32((int32 *)bitmap, fft, count);
		}
	return B_ERROR;
	}

status_t Clear :: Render8(int8 *bitmap, const short *fft, int32 count) {
	int32 width,height;
	width=bounds.IntegerWidth()+1;
	height=bounds.IntegerHeight()+1;

	BScreen screen;
	uint8 index = screen.IndexForColor(
		(color_value>>24)&0xff, (color_value>>16)&0xff, (color_value>>8)&0xff
		);

	if((width*height)&0x7 == 0) {
		uint8 mindex[8];
		for(int32 i=7; i>=0; i--) mindex[i]=index;
		int64 big = *(int64 *)mindex;
		int64 *bit64=(int64 *)bitmap;
		for(int32 i=(width*height>>3)-1; i>=0; i--) bit64[i] = big;
		}
	else memset(bitmap, index, width*height);
	return B_OK;
	}

status_t Clear :: Render16(int16 *bitmap, const short *fft, int32 count) {
	int32 width,height;
	width=bounds.IntegerWidth()+1;
	height=bounds.IntegerHeight()+1;

	int32 r, g, b, color;
	r = (color_value>>24)&0xff;
	g = (color_value>>16)&0xff;
	b = (color_value>>8)&0xff;
	color = ((r&0xf8)<<8) | ((g&0xfc)<<3) | ((b>>3)&0x1f);

	if((width*height)&0x3 == 0) {
		int16 mcolor[4];
		for(int32 i=3; i>=0; i--) mcolor[i]=color;
		int64 big = *(int64 *)mcolor;
		int64 *bit64=(int64 *)bitmap;
		for(int32 i=(width*height>>2)-1; i>=0; i--) bit64[i] = big;
		}
	else for(int32 i=width*height-1; i>=0; i--) bitmap[i] = color;
	return B_OK;
	}

status_t Clear :: Render32(int32 *bitmap, const short *fft, int32 count) {
	int32 width,height;
	width=bounds.IntegerWidth()+1;
	height=bounds.IntegerHeight()+1;

	if((width*height)&0x1 == 0) {
		int32 mcolor[2];
		for(int32 i=1; i>=0; i--) mcolor[i]=color_value;
		int64 big = *(int64 *)mcolor;
		int64 *bit64=(int64 *)bitmap;
		for(int32 i=(width*height>>1)-1; i>=0; i--) bit64[i] = big;
		}
	else for(int32 i=width*height-1; i>=0; i--) bitmap[i] = color_value;
	return B_OK;
	}

status_t Clear :: ScreenModified(BRect bnds, color_space col) {
	bounds=bnds;
	color=col;
	needs_update = true;
	return B_OK;
	}

void Clear :: FileChange(char *name) {
	file_changes++;
	}

void Clear :: MessageReceived(BMessage *message) {
	BView *view=NULL;
	BWindow *win;
	int32 value;
	switch(message->what) {
	case COLOR_MESSAGE:
		message->FindPointer("Color Box", (void **)&view);		
		if(!view) break;
		color_value=((BColorControl *)view)->Value();
		needs_update=true;
		break;
	case CONFIG_MESSAGE:
		message->FindPointer("Config View", (void **)&view);
		if(!view) break;
		win = view->Window();
		if(!win) break;
		if(!win->Lock()) break;
		value=atoi(((BTextControl *)view->FindView("File Change"))->Text());
		if(value<0) value=0;
		if(value != file_change) file_changes=0;
		file_change = value;
		value=atoi(((BTextControl *)view->FindView("Time"))->Text());
		if(value<0) value=0;
		if(value != time) stopwatch->Reset();
		time = value;
		view->RemoveSelf();
		delete view;
		SetHandler(NULL);
		win->Quit();
		break;
	default:
		BLooper::MessageReceived(message);
		break;
		}
	}

int main() {
	BApplication app("application/x-wang-simplescopeplugin");
	BAlert *alert=new BAlert("About Clear",
		"Plugin for SoundPlay.\n"
		"By Joe Wang.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
	}
