
#include <Application.h>
#include <InterfaceKit.h>
#include <SupportDefs.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <StopWatch.h>
#include "fftproto.h"
#include "hold.h"

#define CONFIG_MESSAGE			'CONF'

plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"fft-hold",1, PLUGIN_IS_VISUAL,
	"Hold",
	"By Joe Wang.\n<joe@timewarp.net>\n",
	&GetAbout,
	NULL,
	&GetHoldPlugin,
	&DestroyHoldPlugin
	};

fft_visual_plugin_ops ops={
	{PLUGIN_VISUAL_MAGIC, PLUGIN_VISUAL_VERSION,
		&FileChange, NULL, &GetConfigure, NULL, NULL},
	NULL,
	ScreenModified,
//status_t op_fft_visual_init(); 		(*Init);
	NULL,
	NULL,
	HoldDirect
	};

static plugin_descriptor *plugs[]={
	&plugin,
	0
	};

plugin_descriptor **get_plugin_list(void) {
	return plugs;
	}


class Hold : public BLooper {
private:
	int32 file_changes;
	BRect bounds;
	color_space color;
	bool keeprunning, suspend, needs_update, needs_hold;
	BStopWatch *stopwatch;
	BBitmap *holder8, *holder16, *holder32;
	BHandler *handler;
	
public:
	bigtime_t time;
	int32 file_change;

	Hold();
	~Hold();
	void Quit();
	bool QuitRequested();
	bool SetHandler(BHandler *);
	void MessageReceived(BMessage *mes);
	status_t RenderDirect(void *, const short *fft, int32 count);
	status_t Render8(int8 *, const short *fft, int32 count);
	status_t Render16(int16 *, const short *fft, int32 count);
	status_t Render32(int32 *, const short *fft, int32 count);
	status_t ScreenModified(BRect, color_space);
	void FileChange(char *);
	};

void GetAbout() {
	BAlert *alert=new BAlert("About Hold FFT",
		"Plugin for FFT Plugin.\n"
		"joe@timewarp.net\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
	}

BView *GetConfigure(void *data) {
	Hold *looper = (Hold *)data;
	BView *view=new BView(BRect(0,0,180,80),"Configuration",0,0);
	if(!looper->SetHandler(view)) {
		delete view;
		return NULL;
		}
	view->SetViewColor(40,40,100);
	char *string = (char *)malloc(32);
	sprintf(string, "%d", (int )looper->file_change);
	view->AddChild(
		new BTextControl(BRect(5,5,175,25), "File Change", "# of songs", string, NULL)
		);
	sprintf(string, "%d", (int )looper->time);
	view->AddChild(
		new BTextControl(BRect(5,25,175,45), "Time", "Milliseconds", string, NULL)
		);
	free(string);
	BMessage *msg=NULL;

	msg=new BMessage(CONFIG_MESSAGE);
	msg->AddPointer("Config View", view);
	BButton *button=new BButton(BRect(100,50,150,75), "Apply", "Apply", msg);
	button->SetTarget(NULL, looper);
	view->AddChild(button);
	return view;
	}

void *GetHoldPlugin(void **data,const char *, const char *, uint32, plugin_info *) {
	*data=new Hold();
	return &ops;
	}

void DestroyHoldPlugin(void*,void *data) {
	Hold *looper=(Hold *)data;
	looper->Lock();
	looper->Quit();
	}

status_t HoldDirect(void *data, void *bitmap, const short *buffer, const short *fft, int32 count, void*) {
	Hold *looper=(Hold *)data;
	return looper->RenderDirect(bitmap, fft, count);
	}

status_t ScreenModified(void *data, BRect bounds, color_space col, mini_screen mini, int32 screen) {
	Hold *looper=(Hold *)data;
	return looper->ScreenModified(bounds, col);
	}

void FileChange(void *data, char *name) {
	((Hold *)data)->FileChange(name);
	}

// ==============================================================================
Hold :: Hold() : BLooper() {

	time = 0;
	file_change = file_changes = 0;
	suspend = false;
	needs_update = false;
	needs_hold = true;
	holder8 = holder16 = holder32 = NULL;
	stopwatch = new BStopWatch("timer", true);

	keeprunning=true;
	Run();
	}

Hold :: ~Hold() {
	if(holder8) delete holder8;
	if(holder16) delete holder16;
	if(holder32) delete holder32;
	delete stopwatch;
	}

bool Hold :: QuitRequested() {
	keeprunning=false;
	return false;
	}

void Hold::Quit() {
	keeprunning=false;
	if(handler) {
		BView *view = (BView *)handler;
		if(view->Window() && view->Window()->Lock()) view->Window()->Quit();
		}
	BLooper::Quit();
	}

bool Hold :: SetHandler(BHandler *hand) {
	if(!hand) {
		handler = NULL;
		return true;
		}
	if(handler) {
		BView *view = (BView *)handler;
		if(view->Window()) view->Window()->Show();
		return false;
		}
	handler=hand;
	return true;
	}

status_t Hold :: RenderDirect(void *bitmap, const short *fft, int32 count) {
	if(!keeprunning) return B_ERROR;
	if(time>0) {		//check the timer
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		else if(stopwatch->ElapsedTime()>time*1000) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	if(file_change>0) {
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		if(file_changes >= file_change) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	if(needs_hold) {
		needs_hold = false;
		int32 width, height;
		width = bounds.IntegerWidth()+1;
		height = bounds.IntegerHeight()+1;
		switch(color) {
		case B_CMAP8:
			holder8->SetBits(bitmap, width*height, 0, color);
			holder32->SetBits(bitmap, width*height, 0, color);
			{
				int16 *dest = (int16 *)holder16->Bits();
				int32 *src = (int32 *)holder32->Bits();
				for(int32 i=(holder32->BitsLength()>>2)-1; i>=0; i--) {
					int32 r, g, b;
					r = (src[i]>>16)&0xff;
					g = (src[i]>>8)&0xff;
					b = (src[i]>>0)&0xff;
					dest[i] = ((r&0xf8)<<8) | ((g&0xfc)<<3) | ((b>>3)&0x1f);
					}
				}
			return B_OK;
		case B_RGB16:
			{
				int16 *src = (int16 *)holder16->Bits();
				int16 *src2 = (int16 *)bitmap;
				int32 *dest = (int32 *)holder32->Bits();
				int8 *dest2 = (int8 *)holder8->Bits();
				BScreen screen;
				for(int32 i=(holder32->BitsLength()>>2)-1; i>=0; i--) {
					int32 r, g, b;
					r = (src2[i]>>8)&0xf8;
					g = (src2[i]>>3)&0xfc;
					b = (src2[i]<<3)&0xf8;
					src[i] = src2[i];
					dest[i] = (r<<24) | (g<<16) | (b<<8);
					dest2[i] = screen.IndexForColor(r, g, b);
					}
				}
			return B_OK;
		case B_RGB32:
			holder32->SetBits(bitmap, width*height*4, 0, color);
			{
				int16 *dest = (int16 *)holder16->Bits();
				int8 *dest2 = (int8 *)holder8->Bits();
				int32 *src = (int32 *)holder32->Bits();
				BScreen screen;
				for(int32 i=(holder32->BitsLength()>>2)-1; i>=0; i--) {
					int32 r, g, b;
					r = (src[i]>>24)&0xff;
					g = (src[i]>>16)&0xff;
					b = (src[i]>>8)&0xff;
					dest[i] = ((r&0xf8)<<8) | ((g&0xfc)<<3) | ((b>>3)&0x1f);
					dest2[i] = screen.IndexForColor(r, g, b);
					}
				}
			return B_OK;
		default:
			return B_ERROR;
			}
		return B_ERROR;
		}
	if(!needs_update) return B_OK;

	int32 col=(int32 )color;
	switch(col) {
	case B_CMAP8:
		return Render8((int8 *)bitmap, fft, count);
	case B_RGB16:
		return Render16((int16 *)bitmap, fft, count);
	case B_RGB32:
		return Render32((int32 *)bitmap, fft, count);
	default:
		return B_ERROR;
		}
	return B_ERROR;
	}

status_t Hold :: Render8(int8 *bitmap, const short *fft, int32 count) {
	int32 width, height;
	width = bounds.IntegerWidth()+1;
	height = bounds.IntegerHeight()+1;
	int32 src_width = holder8->Bounds().IntegerWidth()+1;
	int32 scale = src_width*65535/width;

	int8 *bits = (int8 *)holder8->Bits();
	for(int32 i=height-1, j=i*scale; i>=0; i--, j-=scale) {
		int8 *hdest = &bitmap[i*width];
		int8 *hsrc = &bits[(j>>16)*src_width];
		for(int32 k=width-1, l=k*scale; k>=0; k--, l-=scale) {
			hdest[k] = hsrc[l>>16];
			}
		}
	return B_OK;
	}

status_t Hold :: Render16(int16 *bitmap, const short *fft, int32 count) {
	int32 width, height;
	width = bounds.IntegerWidth()+1;
	height = bounds.IntegerHeight()+1;
	int32 src_width = holder16->Bounds().IntegerWidth()+1;
	int32 scale = src_width*65535/width;

	int16 *bits = (int16 *)holder16->Bits();
	for(int32 i=height-1, j=i*scale; i>=0; i--, j-=scale) {
		int16 *hdest = &bitmap[i*width];
		int16 *hsrc = &bits[(j>>16)*src_width];
		for(int32 k=width-1, l=k*scale; k>=0; k--, l-=scale) {
			hdest[k] = hsrc[l>>16];
			}
		}
	return B_OK;
	}

status_t Hold :: Render32(int32 *bitmap, const short *fft, int32 count) {
	int32 width, height;
	width = bounds.IntegerWidth()+1;
	height = bounds.IntegerHeight()+1;
	int32 src_width = holder32->Bounds().IntegerWidth()+1;
	int32 scale = src_width*65535/width;

	int32 *bits = (int32 *)holder32->Bits();
	for(int32 i=height-1, j=i*scale; i>=0; i--, j-=scale) {
		int32 *hdest = &bitmap[i*width];
		int32 *hsrc = &bits[(j>>16)*src_width];
		for(int32 k=width-1, l=k*scale; k>=0; k--, l-=scale) {
			hdest[k] = hsrc[l>>16];
			}
		}
	return B_OK;
	}

status_t Hold :: ScreenModified(BRect bnds, color_space col) {
	bounds=bnds;
	color=col;
	if(needs_hold) {
		if(holder8) delete holder8;
		if(holder16) delete holder16;
		if(holder32) delete holder32;
		holder8 = new BBitmap(bnds, B_CMAP8);
		holder16 = new BBitmap(bnds, B_RGB16);
		holder32 = new BBitmap(bnds, B_RGB32);
		return B_OK;
		}
	needs_update = true;
	return B_OK;
	}

void Hold :: FileChange(char *name) {
	file_changes++;
	}

void Hold :: MessageReceived(BMessage *message) {
	BView *view=NULL;
	BWindow *win;
	int32 value;
	switch(message->what) {
	case CONFIG_MESSAGE:
		message->FindPointer("Config View", (void **)&view);
		if(!view) break;
		win = view->Window();
		if(!win) break;
		if(!win->Lock()) break;
		value=atoi(((BTextControl *)view->FindView("File Change"))->Text());
		if(value<0) value=0;
		if(value != file_change) file_changes=0;
		file_change = value;
		value=atoi(((BTextControl *)view->FindView("Time"))->Text());
		if(value<0) value=0;
		if(value != time) stopwatch->Reset();
		time = value;
		view->RemoveSelf();
		delete view;
		SetHandler(NULL);
		win->Quit();
		break;
	default:
		BLooper::MessageReceived(message);
		break;
		}
	}

int main() {
	BApplication app("application/x-wang-simplescopeplugin");
	BAlert *alert=new BAlert("About Hold",
		"Plugin for SoundPlay.\n"
		"By Joe Wang.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
	}
