
#include <Application.h>
#include <InterfaceKit.h>
#include <SupportDefs.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <StopWatch.h>
#include "fftproto.h"
#include "SimpleScopePlugin.h"

#define COLOR_MESSAGE			'COLR'
#define CONFIG_MESSAGE			'CONF'

plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"fft-simplescope",1, PLUGIN_IS_VISUAL,
	"Simple FFT Scope",
	"By Joe Wang.\n<joe@timewarp.net>\n",
	&getabout,
	NULL,
	&getsimplescopeplugin,
	&destroysimplescopeplugin
	};

fft_visual_plugin_ops ops={
	{PLUGIN_VISUAL_MAGIC, PLUGIN_VISUAL_VERSION,
		&filechange, NULL, &getconfigure, NULL, NULL},
	NULL,
	ScreenModified,
//status_t op_fft_visual_init(); 		(*Init);
	NULL,
	SimpleScope_filter,
	SimpleScope_direct
	};

static plugin_descriptor *plugs[]={
	&plugin,
	0
	};

plugin_descriptor **get_plugin_list(void) {
	return plugs;
	}


class SimpleScopeWindow: public BWindow {
private:
	int32 file_changes;
	BRect bounds;
	color_space color;
	BView *rView;
	BBitmap *bitmap;
	bool keeprunning, suspend;
	BStopWatch *stopwatch;
	BHandler *handler;

public:
	int32 color_value;
	bigtime_t time;
	int32 file_change;

	SimpleScopeWindow();
	~SimpleScopeWindow();
	void Quit();
	bool QuitRequested();
	void MessageReceived(BMessage *mes);
	bool SetHandler(BHandler *);
	status_t Render(const short *fft, int32 count);
	status_t RenderDirect(void *, const short *fft, int32 count);
	status_t Render8(int8 *, const short *fft, int32 count);
	status_t Render16(int16 *, const short *fft, int32 count);
	status_t Render32(int32 *, const short *fft, int32 count);
	status_t ScreenModified(BRect, color_space);
	void FileChange(char *);
	};

void getabout() {
	BAlert *alert=new BAlert("About SimpleScope FFT",
		"SimpleScope plugin for FFT Plugin.\n"
		"joe@timewarp.net\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
	}

BView *getconfigure(void *data) {
	SimpleScopeWindow *window = (SimpleScopeWindow *)data;
	BView *view=new BView(BRect(0,0,280,140),"Configuration",0,0);
	if(!window->SetHandler(view)) {
		delete view;
		return NULL;
		}
	view->SetViewColor(40,40,100);
	char *string = (char *)malloc(32);
	sprintf(string, "%d", (int )window->file_change);
	view->AddChild(
		new BTextControl(BRect(5,5,175,25), "File Change", "# of songs", string, NULL)
		);
	sprintf(string, "%d", (int )window->time);
	view->AddChild(
		new BTextControl(BRect(5,25,175,45), "Time", "Milliseconds", string, NULL)
		);
	free(string);
	BMessage *msg=NULL;
	BColorControl *color=new BColorControl(BPoint(5, 50), B_CELLS_32x8, 0.0F,
		"Color Box");
	color->SetValue(window->color_value);
	msg=new BMessage(COLOR_MESSAGE);
	msg->AddPointer("Color Box", (void *)color);
	color->SetMessage(msg);
	color->SetTarget(NULL, window);
	view->AddChild(color);

	msg=new BMessage(CONFIG_MESSAGE);
	msg->AddPointer("Config View", view);
	BButton *button=new BButton(BRect(100,110,150,135), "Apply", "Apply", msg);
	button->SetTarget(NULL, window);
	view->AddChild(button);
	return view;
	}

void *getsimplescopeplugin(void **data,const char *, const char *, uint32, plugin_info *) {
	*data=new SimpleScopeWindow();
	return &ops;
	}

void destroysimplescopeplugin(void*,void *data) {
	SimpleScopeWindow *win=(SimpleScopeWindow *)data;
	win->Lock();
	win->Quit();
	}

status_t SimpleScope_filter(void *data, const short *buffer, const short *fft, int32 count, void*) {
	SimpleScopeWindow *win=(SimpleScopeWindow *)data;
	if(win->Lock()) {
		win->Show();
		status_t status;
		status=win->Render(fft, count);
		win->Unlock();
		return status;
		}
	return B_ERROR;
	}

status_t SimpleScope_direct(void *data, void *bitmap, const short *buffer, const short *fft, int32 count, void*) {
	SimpleScopeWindow *win=(SimpleScopeWindow *)data;
	if(win->Lock()) {
		status_t status=win->RenderDirect(bitmap, fft, count);
		win->Unlock();
		return status;
		}
	return B_ERROR;
	}

status_t ScreenModified(void *data, BRect bounds, color_space col, mini_screen mini, int32 screen) {
	SimpleScopeWindow *win=(SimpleScopeWindow *)data;
	if(win->Lock()) {
		status_t status=win->ScreenModified(bounds, col);
		win->Unlock();
		return status;
		}
	return B_ERROR;
	}

void filechange(void *data, char *name) {
	((SimpleScopeWindow *)data)->FileChange(name);
	}

// ==============================================================================
SimpleScopeWindow::SimpleScopeWindow()
	: BWindow(BRect(200,100,199+256,99+256),"SimpleScope",B_TITLED_WINDOW,B_NOT_RESIZABLE|B_NOT_ZOOMABLE) {

	time = 0;
	file_change = file_changes = 0;
	suspend = false;
	color_value=0xffffffff;
	handler = NULL;
	stopwatch = new BStopWatch("timer", true);
	rView=new BView(BRect(0,0,256-1,256-1),"view",B_FOLLOW_ALL_SIDES, 0);
	rView->SetViewColor(B_TRANSPARENT_COLOR);
	rView->SetHighColor(255,255,255,255);
	rView->SetFontSize(20);
	AddChild(rView);

	bitmap=new BBitmap(BRect(0,0,256-1,256-1),B_COLOR_8_BIT,false,true);
	Run();
	keeprunning=true;
	}

SimpleScopeWindow::~SimpleScopeWindow() {
	delete stopwatch;
	delete bitmap;
	}

bool SimpleScopeWindow::QuitRequested() {
	keeprunning=false;
	return false;
	}

void SimpleScopeWindow::Quit() {
	keeprunning=false;
	if(handler) {
		BView *view = (BView *)handler;
		if(view->Window() && view->Window()->Lock()) view->Window()->Quit();
		}
	BWindow::Quit();
	}

bool SimpleScopeWindow :: SetHandler(BHandler *hand) {
	if(!hand) {
		handler = NULL;
		return true;
		}
	if(handler) {
		BView *view = (BView *)handler;
		if(view->Window()) view->Window()->Show();
		return false;
		}
	handler=hand;
	return true;
	}

status_t SimpleScopeWindow::Render(const short *fft, int32 count) {
	if(!keeprunning) return B_ERROR;
	if(time>0) {		//check the timer
		if(suspend) {
			suspend=false;
			stopwatch->Reset();
			stopwatch->Resume();
			}
		else if(stopwatch->ElapsedTime()>time*1000) {
			stopwatch->Suspend();
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	if(file_change>0) {
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		if(file_changes >= file_change) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}

	int8 *bits=(int8*)bitmap->Bits();
	memset(bits, 0, 256*256);

	BScreen screen;
	uint8 index = screen.IndexForColor(
		(color_value>>24)&0xff, (color_value>>16)&0xff, (color_value>>8)&0xff
		);

	int32 inc=65536*count/256;
	for(int32 i=0,j=count*65536/2;i<256;i++,j+=inc) {
		if(j>>16 >= count) j=0;
		int32 k=j>>16;
		if(k>=count) k=count-1;
		int32 y=fft[k];
		y*=-256;
		y>>=16;
		y+=128;
		if(y<0) y=0;
		else if(y>256-1) y=256-1;
		bits[i+y*256]= index;
		}

	rView->Sync();
	rView->DrawBitmapAsync(bitmap,bitmap->Bounds(),rView->Bounds());

	return B_OK;
	}

status_t SimpleScopeWindow::RenderDirect(void *bitmap, const short *fft, int32 count) {
	if(!keeprunning) return B_ERROR;
	if(time>0) {		//check the timer
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		else if(stopwatch->ElapsedTime()>time*1000) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	if(file_change>0) {
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		if(file_changes >= file_change) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	int32 col=(int32 )color;
	switch(col) {
	case B_CMAP8:
		return Render8((int8 *)bitmap, fft, count);
	case B_RGB15:
	case B_RGBA15:
	case B_RGB16:
		return Render16((int16 *)bitmap, fft, count);
	case B_RGB32:
		return Render32((int32 *)bitmap, fft, count);
		}
	return B_ERROR;
	}

status_t SimpleScopeWindow::Render8(int8 *bitmap, const short *fft, int32 count) {
	int8 *bits=bitmap;
	int32 width,height;
	width=bounds.IntegerWidth()+1;
	height=bounds.IntegerHeight()+1;
	if((width*height)&0x7 == 0) {
		int64 *bit64=(int64 *)bits;
		for(int32 i=(width*height>>3)-1; i>=0; i--) bit64[i]=0;
		}
	else memset(bits, 0, width*height);

	BScreen screen;
	uint8 index = screen.IndexForColor(
		(color_value>>24)&0xff, (color_value>>16)&0xff, (color_value>>8)&0xff
		);

	bits=bitmap;
	int32 inc=65536*count/width;
	for(int32 i=0,j=count*65536/2;i<width;i++,j+=inc) {
		if(j>>16 > count) j=0;
		int32 y=fft[j>>16];
		y*=-height;
		y>>=16;
		y+=height>>1;
		if(y<0) y=0;
		else if(y>height-1) y=height-1;
		bits[i+y*width]=index;
		}
	return B_OK;
	}

status_t SimpleScopeWindow::Render16(int16 *bitmap, const short *fft, int32 count) {
	int32 width,height;
	width=bounds.IntegerWidth()+1;
	height=bounds.IntegerHeight()+1;

	if((width*height) & 3) {
		int16 *bits=bitmap;
		for(int i=width*height-1; i>=0;i--) *bits++=0;
		}
	else {
		int64 *bits=(int64 *)bitmap;
		for(int i=((width*height)>>2)-1; i>=0; i--) *bits++=0;
		}

	int32 r, g, b, color;
	r = (color_value>>24)&0xff;
	g = (color_value>>16)&0xff;
	b = (color_value>>8)&0xff;
	color = ((r&0xf8)<<8) | ((g&0xfc)<<3) | ((b>>3)&0x1f);

	int16 *bits=bitmap;
	int32 inc=65536*count/width;
	for(int32 i=0,j=count*65536/2;i<width;i++,j+=inc) {
		if(j>>16 > count) j=0;
		int32 y=fft[j>>16];
		y*=-height;
		y>>=16;
		y+=height>>1;
		if(y<0) y=0;
		else if(y>height-1) y=height-1;
		bits[i+y*width]=color;
		}
	return B_OK;
	}

status_t SimpleScopeWindow::Render32(int32 *bitmap, const short *fft, int32 count) {
	int32 *bits=bitmap;
	int32 width,height;
	width=bounds.IntegerWidth()+1;
	height=bounds.IntegerHeight()+1;
	for(int i=0;i<width*height;i++) *bits++=0;

	bits=bitmap;
	int32 inc=65536*count/width;
	for(int32 i=0,j=count*65536/2;i<width;i++,j+=inc) {
		if(j>>16 > count) j=inc;
		int32 y=fft[j>>16];
		y*=-height;
		y>>=16;
		y+=height>>1;
		if(y<0) y=0;
		else if(y>height-1) y=height-1;
		bits[i+y*width]=color_value;
		}
	return B_OK;
	}

status_t SimpleScopeWindow::ScreenModified(BRect bnds, color_space col) {
	bounds=bnds;
	color=col;
	return B_OK;
	}

void SimpleScopeWindow :: FileChange(char *name) {
	file_changes++;
	SetTitle(name);
	}

void SimpleScopeWindow::MessageReceived(BMessage *message) {
	BView *view=NULL;
	BWindow *win;
	int32 value;
	switch(message->what) {
	case COLOR_MESSAGE:
		message->FindPointer("Color Box", (void **)&view);		
		if(!view) break;
		color_value=((BColorControl *)view)->Value();
		break;
	case CONFIG_MESSAGE:
		message->FindPointer("Config View", (void **)&view);
		if(!view) break;
		win = view->Window();
		if(!win) break;
		if(!win->Lock()) break;
		value=atoi(((BTextControl *)view->FindView("File Change"))->Text());
		if(value<0) value=0;
		if(value != file_change) file_changes=0;
		file_change = value;
		value=atoi(((BTextControl *)view->FindView("Time"))->Text());
		if(value<0) value=0;
		if(value != time) stopwatch->Reset();
		time = value;
		view->RemoveSelf();
		delete view;
		SetHandler(NULL);
		win->Quit();
		break;
	default:
		BWindow::MessageReceived(message);
		break;
		}
	}

int main() {
	BApplication app("application/x-wang-simplescopeplugin");
	BAlert *alert=new BAlert("About SimpleScope FFT",
		"SimpleScope plugin for SoundPlay.\n"
		"By Joe Wang modified from Marco Nelissen's Original source.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
	}
