
#include <Application.h>
#include <InterfaceKit.h>
#include <SupportDefs.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <StopWatch.h>
#include "fftproto.h"
#include "strobe.h"

#define COLOR_MESSAGE			'COLR'
#define CONFIG_MESSAGE			'CONF'

static plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"fft-diff-strobe",1, PLUGIN_IS_VISUAL,
	"Differential Strobe",
	"By Joe Wang.\n<joe@timewarp.net>\n",
	&getabout,
	NULL,
	&getstrobeplugin,
	&destroystrobeplugin
	};

fft_visual_plugin_ops ops={
	{PLUGIN_VISUAL_MAGIC, PLUGIN_VISUAL_VERSION,
		&filechange, NULL, &getconfigure, NULL, NULL},
	NULL,
	ScreenModified,
//status_t op_fft_visual_init(); 		(*Init);
	NULL,
	Strobe_filter,
	Strobe_direct
	};

plugin_descriptor *plugs[]={
	&plugin,
	0
	};

plugin_descriptor **get_plugin_list(void) {
	return plugs;
	}


class StrobeWindow: public BWindow {
private:
	int32 file_changes, previous, delta;
	BRect bounds;
	color_space color;
	BView *rView;
	BBitmap *bitmap;
	bool keeprunning, suspend;
	BStopWatch *stopwatch;
	BHandler *handler;

public:
	int32 color_value;
	bigtime_t time;
	int32 file_change;

	StrobeWindow();
	~StrobeWindow();
	void Quit();
	bool QuitRequested();
	void MessageReceived(BMessage *mes);
	bool SetHandler(BHandler *);
	status_t Render(const short *fft, int32 count);
	status_t RenderDirect(void *, const short *fft, int32 count);
	status_t Render8(int8 *, const short *fft, int32 count);
	status_t Render16(int16 *, const short *fft, int32 count);
	status_t Render32(int32 *, const short *fft, int32 count);
	status_t ScreenModified(BRect, color_space);
	void FileChange(char *);
	};

void getabout() {
	BAlert *alert=new BAlert("About Differential Strobe FFT",
		"Differential Strobe plugin for FFT Plugin.\n"
		"joe@timewarp.net\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
	}

BView *getconfigure(void *data) {
	StrobeWindow *window = (StrobeWindow *)data;
	BView *view=new BView(BRect(0,0,280,140),"Configuration",0,0);
	if(!window->SetHandler(view)) {
		delete view;
		return NULL;
		}
	view->SetViewColor(40,40,100);
	char *string = (char *)malloc(32);
	sprintf(string, "%d", (int )window->file_change);
	view->AddChild(
		new BTextControl(BRect(5,5,175,25), "File Change", "# of songs", string, NULL)
		);
	sprintf(string, "%d", (int )window->time);
	view->AddChild(
		new BTextControl(BRect(5,25,175,45), "Time", "Milliseconds", string, NULL)
		);
	free(string);
	BMessage *msg=NULL;
	BColorControl *color=new BColorControl(BPoint(5, 50), B_CELLS_32x8, 0.0F,
		"Color Box");
	color->SetValue(window->color_value);
	msg=new BMessage(COLOR_MESSAGE);
	msg->AddPointer("Color Box", (void *)color);
	color->SetMessage(msg);
	color->SetTarget(NULL, window);
	view->AddChild(color);

	msg=new BMessage(CONFIG_MESSAGE);
	msg->AddPointer("Config View", view);
	BButton *button=new BButton(BRect(100,110,150,135), "Apply", "Apply", msg);
	button->SetTarget(NULL, window);
	view->AddChild(button);
	return view;
	}

void *getstrobeplugin(void **data,const char *, const char *, uint32, plugin_info *) {
	*data=new StrobeWindow();
	return &ops;
	}

void destroystrobeplugin(void*,void *data) {
	StrobeWindow *win=(StrobeWindow *)data;
	win->Lock();
	win->Quit();
	}

status_t Strobe_filter(void *data, const short *buffer, const short *fft, int32 count, void*) {
	StrobeWindow *win=(StrobeWindow *)data;
	if(win->Lock()) {
		if(win->IsHidden()) win->Show();
		status_t status;
		status=win->Render(fft, count);
		win->Unlock();
		return status;
		}
	return B_ERROR;
	}

static status_t Strobe_direct(void *data, void *bitmap, const short *buffer, const short *fft, int32 count, void*) {
	StrobeWindow *win=(StrobeWindow *)data;
//	if(win->Lock()) {
		status_t status=win->RenderDirect(bitmap, fft, count);
//		win->Unlock();
		return status;
//		}
	return B_ERROR;
	}

status_t ScreenModified(void *data, BRect bounds, color_space col, mini_screen mini, int32 screen) {
	StrobeWindow *win=(StrobeWindow *)data;
	if(win->Lock()) {
		status_t status=win->ScreenModified(bounds, col);
		win->Unlock();
		return status;
		}
	return B_ERROR;
	}

void filechange(void *data, char *name) {
	((StrobeWindow *)data)->FileChange(name);
	}

// ==============================================================================
StrobeWindow::StrobeWindow()
	: BWindow(BRect(200,100,199+256,99+256),"Strobe",B_TITLED_WINDOW,B_NOT_RESIZABLE|B_NOT_ZOOMABLE) {

	time = 0;
	file_change = file_changes = previous = delta = 0;
	suspend = false;
	color_value=0xffffffff;
	handler = NULL;
	stopwatch = new BStopWatch("timer", true);
	rView=new BView(BRect(0,0,256-1,256-1),"view",B_FOLLOW_ALL_SIDES, 0);
	rView->SetViewColor(B_TRANSPARENT_COLOR);
	rView->SetHighColor(255,255,255,255);
	rView->SetFontSize(20);
	AddChild(rView);

	bitmap=new BBitmap(BRect(0,0,256-1,256-1),B_COLOR_8_BIT,false,true);

	keeprunning=true;
	Run();
	}

StrobeWindow::~StrobeWindow() {
	delete stopwatch;
	delete bitmap;
	}

bool StrobeWindow::QuitRequested() {
	keeprunning=false;
	return false;
	}

void StrobeWindow::Quit() {
	keeprunning=false;
	if(handler) {
		BView *view = (BView *)handler;
		if(view->Window() && view->Window()->Lock()) view->Window()->Quit();
		}
	BWindow::Quit();
	}

bool StrobeWindow :: SetHandler(BHandler *hand) {
	if(!hand) {
		handler = NULL;
		return true;
		}
	if(handler) {
		BView *view = (BView *)handler;
		if(view->Window()) view->Window()->Show();
		return false;
		}
	handler=hand;
	return true;
	}

status_t StrobeWindow::Render(const short *fft, int32 count) {
	if(!keeprunning) return B_ERROR;
	if(time>0) {		//check the timer
		if(suspend) {
			suspend=false;
			stopwatch->Reset();
			stopwatch->Resume();
			}
		else if(stopwatch->ElapsedTime()>time*1000) {
			stopwatch->Suspend();
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	if(file_change>0) {
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		if(file_changes >= file_change) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}

	int32 value = 0;
	for(int32 i=0; i<count; i++) {
		int16 impulse = fft[i];
		if(impulse<0) impulse *= -1;
		value+=impulse;
		}
	value /= count;
	
	delta = (value-previous)>>1;
	previous = value;
	
	if(delta<0) delta=0;
	else if(delta>255) delta=255;

	int32 c = color_value>>8;
	int32 r_b = c&0xff00ff;
	int32 g = c&0xff00;
	r_b *= delta;
	g *= delta;

	BScreen screen;
	uint8 index = screen.IndexForColor(
		(r_b>>24)&0xff, (g>>16)&0xff, (r_b>>8)&0xff
		);

	int8 mindex[8];
	for(int32 i=7; i>=0; i--) mindex[i]=index;
	int64 big = *(int64 *)mindex;
	
	int64 *bits=(int64 *)bitmap->Bits();

	rView->Sync();
	for(int32 i=65536/8-1; i>=0; i--) *bits++=big;

	rView->DrawBitmapAsync(bitmap,bitmap->Bounds(),rView->Bounds());

	return B_OK;
	}

status_t StrobeWindow::RenderDirect(void *bitmap, const short *fft, int32 count) {
	if(!keeprunning) return B_ERROR;
	if(time>0) {		//check the timer
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		else if(stopwatch->ElapsedTime()>time*1000) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	if(file_change>0) {
		if(suspend) {
			suspend=false;
			file_changes=0;
			stopwatch->Reset();
			}
		if(file_changes >= file_change) {
			suspend=true;
			return B_TIMED_OUT;
			}
		}
	int32 value = 0;
	for(int32 i=0; i<count; i++) {
		int16 impulse = fft[i];
		if(impulse<0) impulse *= -1;
		value+=impulse;
		}
	value /= count;
	
	delta = (value-previous)>>1;
	previous = value;
	
	if(delta<0) delta=0;
	else if(delta>255) delta=255;

	int32 col=(int32 )color;
	switch(col) {
	case B_CMAP8:
		return Render8((int8 *)bitmap, fft, count);
	case B_RGB15:
	case B_RGBA15:
	case B_RGB16:
		return Render16((int16 *)bitmap, fft, count);
	case B_RGB32:
		return Render32((int32 *)bitmap, fft, count);
		}
	return B_ERROR;
	}

status_t StrobeWindow::Render8(int8 *bitmap, const short *fft, int32 count) {
	int32 width,height;
	width=bounds.IntegerWidth()+1;
	height=bounds.IntegerHeight()+1;

	int32 c = color_value>>8;
	int32 r_b = c&0xff00ff;
	int32 g = c&0xff00;
	r_b *= delta;
	g *= delta;

	BScreen screen;
	uint8 index = screen.IndexForColor(
		(r_b>>24)&0xff, (g>>16)&0xff, (r_b>>8)&0xff
		);

	if((width*height)&0x7 == 0) {
		uint8 mindex[8];
		for(int32 i=7; i>=0; i--) mindex[i]=index;
		int64 big = *(int64 *)mindex;
		int64 *bit64=(int64 *)bitmap;
		for(int32 i=(width*height>>3)-1; i>=0; i--) bit64[i] = big;
		}
	else memset(bitmap, index, width*height);
	return B_OK;
	}

status_t StrobeWindow::Render16(int16 *bitmap, const short *fft, int32 count) {
	int32 width,height;
	width=bounds.IntegerWidth()+1;
	height=bounds.IntegerHeight()+1;

	int32 c = color_value>>8;
	int32 r_b = c&0xff00ff;
	int32 g = c&0xff00;
	r_b *= delta;
	g *= delta;

	int32 r, b, color;
	r = (r_b>>24)&0xff;
	g = (g>>16)&0xff;
	b = (r_b>>8)&0xff;
	color = ((r&0xf8)<<8) | ((g&0xfc)<<3) | ((b>>3)&0x1f);

	if((width*height)&0x3 == 0) {
		int16 mcolor[4];
		for(int32 i=3; i>=0; i--) mcolor[i]=color;
		int64 big = *(int64 *)mcolor;
		int64 *bit64=(int64 *)bitmap;
		for(int32 i=(width*height>>2)-1; i>=0; i--) bit64[i] = big;
		}
	else for(int32 i=width*height-1; i>=0; i--) bitmap[i] = color;
	return B_OK;
	}

status_t StrobeWindow::Render32(int32 *bitmap, const short *fft, int32 count) {
	int32 width,height;
	width=bounds.IntegerWidth()+1;
	height=bounds.IntegerHeight()+1;

	int32 c = color_value>>8;
	int32 r_b = c&0xff00ff;
	int32 g = c&0xff00;
	r_b *= delta;
	g *= delta;

	int32 color = (r_b&0xff00ff00)+(g&0xff0000);

	if((width*height)&0x1 == 0) {
		int32 mcolor[2];
		for(int32 i=1; i>=0; i--) mcolor[i]=color;
		int64 big = *(int64 *)mcolor;
		int64 *bit64=(int64 *)bitmap;
		for(int32 i=(width*height>>1)-1; i>=0; i--) bit64[i] = big;
		}
	else for(int32 i=width*height-1; i>=0; i--) bitmap[i] = color;
	return B_OK;
	}

status_t StrobeWindow::ScreenModified(BRect bnds, color_space col) {
	bounds=bnds;
	color=col;
	return B_OK;
	}

void StrobeWindow :: FileChange(char *name) {
	file_changes++;
	SetTitle(name);
	}

void StrobeWindow::MessageReceived(BMessage *message) {
	BView *view=NULL;
	BWindow *win;
	int32 value;
	switch(message->what) {
	case COLOR_MESSAGE:
		message->FindPointer("Color Box", (void **)&view);		
		if(!view) break;
		color_value=((BColorControl *)view)->Value();
		break;
	case CONFIG_MESSAGE:
		message->FindPointer("Config View", (void **)&view);
		if(!view) break;
		win = view->Window();
		if(!win) break;
		if(!win->Lock()) break;
		value=atoi(((BTextControl *)view->FindView("File Change"))->Text());
		if(value<0) value=0;
		if(value != file_change) file_changes=0;
		file_change = value;
		value=atoi(((BTextControl *)view->FindView("Time"))->Text());
		if(value<0) value=0;
		if(value != time) stopwatch->Reset();
		time = value;
		view->RemoveSelf();
		delete view;
		SetHandler(NULL);
		win->Quit();
		break;
	default:
		BWindow::MessageReceived(message);
		break;
		}
	}

int main() {
	BApplication app("application/x-wang-simplescopeplugin");
	BAlert *alert=new BAlert("About Differential Strobe",
		"Differential Strobe plugin.\n"
		"By Joe Wang.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
	}
